# Sample script to test the Sparse Bayesian Linear Regression algorithm
#
# 2013 written by Ralf Herbrich

# Divides a message out of a Gaussian 
divideOut <- function(gauss, gaussMsg) {
  tau = gauss$mu / gauss$sigma2 - gaussMsg$tau;
  pi = 1 / gauss$sigma2 - gaussMsg$pi;
  return (list(mu = tau/pi, sigma2 = 1/pi));
}

# Multiplies a message into a Gaussian 
multiplyIn <- function(gauss, gaussMsg) {
  tau = gauss$mu / gauss$sigma2 + gaussMsg$tau;
  pi = 1 / gauss$sigma2 + gaussMsg$pi;
  return (list(mu = tau/pi, sigma2 = 1/pi));a
}

# Divides a message out of a Gaussian 
divide <- function(gauss1, gauss2) {
  tau = gauss1$mu / gauss1$sigma2 - gauss2$mu / gauss2$sigma2;
  pi = 1 / gauss1$sigma2 - 1 / gauss2$sigma2;
  return (list(tau = tau, pi = pi));
}

# Compute the Kullback-Leibler divergence of two Gaussians
klDivergence <- function(gauss1, gauss2) {
  return(0.5*(gauss1$sigma2/gauss2$sigma2 + 
                (gauss1$mu - gauss2$mu)^2/gauss2$sigma2 - 
                log(gauss1$sigma2/gauss2$sigma2) - 1));
}

# Converts a factorized Gaussian belief into a multivariate Gaussian
convertBelief <- function(wBelief) {
  return(
    list(mu = c(wBelief[[1]]$mu, wBelief[[2]]$mu),
         Sigma = matrix(c(wBelief[[1]]$sigma2,0,0,wBelief[[2]]$sigma2),2))
    );
}

# Computes the update of the marginal as a function of the x-vector and y (2D)
update2D <- function(w,x,y,beta2) {
  s = x[1]*x[1]*w[[1]]$sigma2 + x[2]*x[2]*w[[2]]$sigma2 + beta2;
  delta = y - (x[1]*w[[1]]$mu + x[2]*w[[2]]$mu);
  w1_mu = w[[1]]$mu + x[1]*w[[1]]$sigma2/s * delta;  
  w2_mu = w[[2]]$mu + x[2]*w[[2]]$sigma2/s * delta;
  w1_sigma2 = w[[1]]$sigma2*(1 - x[1]*x[1]*w[[1]]$sigma2/s);
  w2_sigma2 = w[[2]]$sigma2*(1 - x[2]*x[2]*w[[2]]$sigma2/s);
  w1 = list(mu=w1_mu,sigma2=w1_sigma2);
  w2 = list(mu=w2_mu,sigma2=w2_sigma2);
  return (list(w1, w2));
}

# Compute the best posterior using the EP algorithm
EP2D <- function(priorW,X,y,beta2, epsilon=1e-16, quiet=FALSE) {
  # Initialize
  q = priorW;
  msg = list();
  for (i in 1:nrow(X)) {
    msg[[i]] = list(list(tau=0, pi=0),list(tau=0, pi=0));
  }
  
  # EP loop
  delta = 1e20;
  iter = 1;
  while(delta > epsilon) {
    # Initialize the delta posterior
    delta = 0;
    # Refine the message for each likleihood term
    for (i in 1:nrow(X)) {
      # (1) Remove the messages from the current marginal
      p = list(divideOut(q[[1]],msg[[i]][[1]]), 
               divideOut(q[[2]],msg[[i]][[2]]));
      
      # (2) Compute the best approximation using a factorized belief distribution over weights
      qNew = update2D(p, X[i,], y[i], beta2);
      
      # (3) Update the messages for this factor
      msg[[i]] = list(divide(qNew[[1]], p[[1]]),
                      divide(qNew[[2]], p[[2]]));
      
      # (4) Update the delta
      delta = max(delta, 
                  klDivergence(qNew[[1]],q[[1]]),
                  klDivergence(qNew[[2]],q[[2]]));
      
      # (5) Assign the new posterior
      q = qNew;
    }
    if (!quiet) {
      print(paste('Delta in Iteration ', iter, ' = ', delta), sep="");       
    }
    iter = iter + 1;
  }
  
  return(q);
}

# Compute the best posterior using the EP algorithm in batches using MapReduce
EP2DMapReduce <- function(priorW,X,y,beta2, epsilonEP=1e-16, epsilonMR=1e-16, noMappers=4) {
  # First, work out the chunks on which we run MapReduce
  mapIdx = split(1:nrow(X),factor(floor(noMappers*runif(nrow(X)))));
  
  # Initialize the messages and posterior
  msgs = Map(function(idx) { return(list(list(tau=0,pi=0),list(tau=0,pi=0))) }, mapIdx);
  q = priorW;

  delta = 1e20;
  iter = 1;
  while(delta > epsilonMR) {
    msgsNew = Map(function(idx, msg) {
      # select training subsets
      trainX = X[idx,,drop=FALSE];
      trainy = y[idx];
      
      # (1) Remove the messages from the current best marginal (from last MapReduce run)
      prior = list(divideOut(q[[1]],msg[[1]]), 
                   divideOut(q[[2]],msg[[2]]));
      
      # (2) run EP to get the posterior
      posterior = EP2D(prior,trainX,trainy,beta2,epsilonEP,quiet=TRUE);
      
      # (3) compute the new messages for this training block
      return (list(divide(posterior[[1]], prior[[1]]),
                   divide(posterior[[2]], prior[[2]])));
    }, mapIdx, msgs);
    
    qNew = Reduce(function(post, msg) { 
      return(list(multiplyIn(post[[1]],msg[[1]]),
                  multiplyIn(post[[2]],msg[[2]]))) 
      }, msgsNew, priorW); 
    
    # (4) Update the delta
    delta = max(klDivergence(qNew[[1]],q[[1]]),
                klDivergence(qNew[[2]],q[[2]]));
    print(paste('Delta in MapReduce Iteration ', iter, ' = ', delta), sep="");       
    iter = iter + 1;
    
    # (5) Update the posterior and messages
    q = qNew;
    msgs = msgsNew;
  }
  
  return(qNew);
}

# Compute the best posterior using the EP algorithm
BLR2D <- function(priorW,X,y,beta2) {
  prior = convertBelief(priorW);
  Sigma = solve(solve(prior$Sigma)+1/beta2*t(X)%*%X);
  mu = Sigma %*% (solve(prior$Sigma)%*%prior$mu + 1/beta2*t(X)%*%y);
  return(list(mu = mu, Sigma = Sigma));
}

# Setup the data for the sample 2D problem
setup2D <- function(from=-3, to=-3, plotN=1000) {
  # Setup the input data
  inp <- seq(from=-from,to=to,length.out=plotN);
  X = cbind(sin(3*inp),sin(inp)*cos(6*inp));
  return (list(inp = inp, X = X))
}

# Draw the function for a given weight vector (2D problem) 
drawFunction2D <- function(plotData, w, lwd=1, lty=1, col=rgb(red=0,green=0,blue=0)) {
  y = plotData$X %*% w;
  lines(plotData$inp,y,lwd=lwd,lty=lty,col=col);  
}

# Draw the function for a given weight belief (2D problem)
drawFunctionWithUncertainty2D <- function(plotData, wBelief, beta2, 
                                          lwd=4, lty=1, col=rgb(red=1,green=0,blue=0)) {
  mu = c(wBelief[[1]]$mu, wBelief[[2]]$mu);
  sigma2 = c(wBelief[[1]]$sigma2, wBelief[[2]]$sigma2);
  mean = plotData$X %*% mu;
  variance = plotData$X^2 %*% sigma2 + beta2;
  xx = c(plotData$inp, rev(plotData$inp));
  yy = c(mean+sqrt(variance), rev(mean-sqrt(variance)));

  c = col2rgb(col);  
  lines(plotData$inp,mean,lwd=lwd,lty=lty,col=col);
  polygon(xx,yy, col=rgb(red=c[1]/255, green=c[2]/255, blue=c[3]/255, alpha=0.2));
}

# Computes the density of the 2D Gaussian
dnorm2d <- function(x,y,mu,Sigma) {
  SigmaInv = solve(Sigma);
  res = x;
  c = 1/(2*pi*sqrt(det(Sigma)));
  for (i in 1:length(x)) {
    d = c(x[i],y[i]) - mu;
    res[i] = c*exp(-0.5*(t(d)%*%SigmaInv%*%d));
  }
  return(res);
}

# Plots the contour plot of a multivariate Gaussian
plotBelief <- function(mvGaussian, N=100, levels=NA, ...) {
  x = seq(from=mvGaussian$mu[1]-3*sqrt(mvGaussian$Sigma[1,1]),
          to=mvGaussian$mu[1]+3*sqrt(mvGaussian$Sigma[1,1]),
          length.out=N);
  y = seq(from=mvGaussian$mu[2]-3*sqrt(mvGaussian$Sigma[2,2]),
          to=mvGaussian$mu[2]+3*sqrt(mvGaussian$Sigma[2,2]),
          length.out=N);
  z = outer(x,y,FUN=dnorm2d, mu=mvGaussian$mu, Sigma=mvGaussian$Sigma);
  if (length(levels) == 1 && is.na(levels)) {
    levels = pretty(range(z,finite=TRUE), 10);    
  }
  contour(x,y,z, nlevels=10, levels=levels, ...);
  return(levels);
}

# 2D demo of the algorithm
demo2D <- function(noSampleWeights=20, noTrainSamples=10, beta2=0.1, noMappers=4) {
  # Setup the demo
  plotData = setup2D();
  set.seed(1508);

  # Plot the prior functions
  plot(range(plotData$inp),c(-3,+3),pch='',xlab='Input',ylab='Output');
  title("Sample Functions from the Prior")
  for (i in 1:noSampleWeights) {
    w = rnorm(2,mean=0,sd=1);
    drawFunction2D(plotData,w);
  }
  readline(prompt="Press <ENTER> for next plot")  
  
  # Draw a sample weight to learn from and create the training set
  set.seed(1234);
  plot.new();
  plot(range(plotData$inp),c(-3,+3),type="n",xlab='Input',ylab='Output');
  title("Sample Function and Training Data");
  
  wStar = rnorm(2,mean=0,sd=1);
  drawFunction2D(plotData,wStar,lwd=4);
  trainIdx = sample.int(n=length(plotData$inp), size=noTrainSamples, replace=FALSE);
  trainX = plotData$X[trainIdx,];
  trainY = trainX %*% wStar + rnorm(n=noTrainSamples,mean=0,sd=sqrt(beta2));
  points(plotData$inp[trainIdx],trainY,pch='x');

  wPrior = list(list(mu=0,sigma2=1),list(mu=0,sigma2=1));
  wPosterior = EP2D(wPrior,trainX,trainY,beta2);
  drawFunctionWithUncertainty2D(plotData,wPosterior,beta2);  
  readline(prompt="Press <ENTER> for next plot");
  
  wPrior = list(list(mu=0,sigma2=1),list(mu=0,sigma2=1));
  wPosteriorMR = EP2DMapReduce(wPrior,trainX,trainY,beta2,noMappers=noMappers);
  drawFunctionWithUncertainty2D(plotData,wPosteriorMR,beta2,col="green");  
  readline(prompt="Press <ENTER> for next plot");
  
  wTruePosterior = BLR2D(wPrior,trainX,trainY,beta2);
  wApproxPosterior = convertBelief(wPosterior);
  title("Posterior Density Plot");
  levels = plotBelief(wTruePosterior);
  levels = plotBelief(wApproxPosterior, levels=levels, col="red", add=TRUE);  
  print(paste ("True Posterior Mean = ", wTruePosterior$mu));
  print(paste ("Approximate Mean = ", wApproxPosterior$mu));
}

demo2D()



